package kernel.net;

import kernel.log.Log;
import kernel.peripherals.Peripherals.Peripheral;
import kernel.net.INetworkInterface;
import kernel.net.Package.NetworkID;

using tink.CoreApi;

@:structInit typedef Route = {
	interf:INetworkInterface,
	rep:NetworkID,
	cost:Int
}

/**
	Manages network routing.
**/
class Routing {
	/**
		Depends on: Peripheral
	**/
	public static var instance:Routing;

	public static inline final UPDATE_WAIT_TIME:Float = 1;

	public final onNewNeigbor:Signal<Int>;

	private final onNewNeigborTrigger:SignalTrigger<NetworkID> = Signal.trigger();
	private final routingTable:Map<NetworkID, Route> = new Map();
	private var routeUpdateInQueue:Bool = false;

	@:allow(kernel.Init)
	private function new() {
		this.onNewNeigbor = this.onNewNeigborTrigger.asSignal();
	}

	@:allow(kernel.Init)
	private function init() {
		routingTable.set(Net.instance.networkID, {interf: Loopback.instance, cost: 0, rep: Net.instance.networkID});
		brodcastRoutingTable();
	}

	/**
		Prepares an brodcast of the current routing table. If a brodcast is already in queue it will be ignored.
		After UPDATE_WAIT_TIME seconds the routing table will be brodcasted. This is done to prevent spamming the network.
	**/
	private function prepareRouteUpdate() {
		if (this.routeUpdateInQueue) {
			return;
		}

		this.routeUpdateInQueue = true;

		new Timer(UPDATE_WAIT_TIME, () -> {
			brodcastRoutingTable();
			this.routeUpdateInQueue = false;
		});
	}

	/**
		Brodcast the current routing table to all peers.
	**/
	private function brodcastRoutingTable() {
		var pack = newRoutDiscoverPackage();

		for (modem in Peripheral.instance.getAllModems()) {
			modem.send(Net.BRODCAST_PORT, Net.instance.networkID, pack);
		}
	}

	/**
		Handle incomming packages that involve route discovery.
	**/
	@:allow(kernel.net.Net)
	private function handleRoutePackage(pack:Package<Noise>, interf:INetworkInterface):Void {
		addPossibleRoute(pack.fromID, interf, 0, pack.fromID);

		var shouldRespond:Bool = switch pack.type {
			case RouteDiscoverResponse(routes):
				for (route in routes) {
					addPossibleRoute(route.id, interf, route.cost, pack.fromID);
				}
				false;
			case RouteDiscover(routes):
				for (route in routes) {
					addPossibleRoute(route.id, interf, route.cost, pack.fromID);
				}
				true;
			case RouteDiscoverUpdate(routes):
				for (route in routes) {
					addPossibleRoute(route.id, interf, route.cost, pack.fromID);
				}
				false;
			default:
				Log.error("Expected package to be a Route discover package");
				false;
		};

		if (!shouldRespond) {
			return;
		}

		// Respond to peer
		var response:Package<Noise> = {
			toID: pack.fromID,
			fromID: Net.instance.networkID,
			msgID: null,
			type: RouteDiscoverResponse(genRouteList()),
			data: null,
			ttl: 0, // Prevent forwarding
		}

		interf.send(response.toID, Net.instance.networkID, response);
	}

	/**
		Generate a list of routes to send to a peer based on the current routing table.
	**/
	private function genRouteList():Array<{id:NetworkID, cost:Int}> {
		var routes:Array<{id:NetworkID, cost:Int}> = [];
		for (k => v in this.routingTable) {
			routes.push({
				id: k,
				cost: v.cost
			});
		}
		return routes;
	}

	private function newRoutDiscoverPackage():Package<Noise> {
		var pack:Package<Noise> = {
			type: RouteDiscover(genRouteList()),
			toID: Net.BRODCAST_PORT,
			msgID: null,
			fromID: Net.instance.networkID,
			data: null,
			ttl: 0, // Prevent forwarding
		}

		return pack;
	}

	/**
		Called when a route to a client has been disoverd.
		Its possible to be called multiple times with the same id but different addr.
	**/
	private function addPossibleRoute(toID:Int, interf:INetworkInterface, cost:Int, rep:NetworkID) {
		if (toID == Net.instance.networkID) {
			return;
		}

		var fullCost = cost + interf.getBaseRoutingCost();

		if (this.routingTable.exists(toID)) {
			if (this.routingTable[toID].cost > fullCost) {
				this.routingTable[toID] = {interf: interf, cost: cost + interf.getBaseRoutingCost(), rep: rep};
				this.prepareRouteUpdate();
			}
		} else {
			this.routingTable[toID] = {interf: interf, cost: cost + interf.getBaseRoutingCost(), rep: rep};
			this.onNewNeigborTrigger.trigger(toID);
			this.prepareRouteUpdate();
		}
	}

	public function getRouteToID(networkID:NetworkID):{interf:INetworkInterface, rep:NetworkID} {
		if (networkID == Net.instance.networkID) {
			return {interf: Loopback.instance, rep: Net.instance.networkID};
		}

		var route:Null<Route> = this.routingTable[networkID];

		if (route == null) {
			return null;
		} else {
			return {interf: route.interf, rep: route.rep};
		}
	}

	public function getRouteTable():Map<NetworkID, Route> {
		return this.routingTable;
	}
}
